import os
from dataclasses import dataclass, fields

import yaml

from dagster_uc.log import logger


@dataclass
class UserCodeDeploymentsConfig:
    """This is a data class that holds the configuration parameters necessary
    for running the user code deployment script.
    """

    environment: str
    container_registry: str
    dockerfile: str
    image_prefix: str
    namespace: str
    node: str
    code_path: str
    docker_root: str
    repository_root: str
    dagster_version: str
    user_code_deployment_env_secrets: list[
        dict[str, str]
    ]  # Must be list of dicts with key 'name' like so: [{"name": "sp-credentials"}, {"name": "lakefs-credentials"}]
    user_code_deployment_env: list[
        dict[str, str]
    ]  # Must be list of dicts with keys 'name' and 'value' like so: [{"name": "MY_ENV_VAR", "value": "True"}, ...]
    cicd: bool
    limits: dict[str, str]
    requests: dict[str, str]
    kubernetes_context: str
    dagster_gui_url: str | None = None
    verbose: bool = False
    use_az_login: bool = True
    use_project_name: bool = True
    user_code_deployments_configmap_name: str = "dagster-user-deployments-values-yaml"
    dagster_workspace_yaml_configmap_name: str = "dagster-workspace-yaml"
    uc_deployment_semaphore_name: str = "dagster-uc-semaphore"


def load_config(environment: str, path: str | None) -> "UserCodeDeploymentsConfig":
    """Loads the configuration file from the local dir or the user's home dir."""
    if path is None:
        paths_to_try = [
            ".config_user_code_deployments.yaml",
            "~/.config_user_code_deployments.yaml",
        ]
        for path_to_try in paths_to_try:
            path_to_try = os.path.expanduser(path_to_try)
            if os.path.exists(path_to_try):
                path = path_to_try
        if path is None:
            raise Exception(
                f"Could not load config file. Tried the following locations: {paths_to_try}\nCurrent folder: {os.getcwd()}\nContents of current folder: {os.listdir(os.getcwd())}\n\n Tip: Place the config file in the same folder you're calling this script from, or your home directory, or specify the path manually using --config-file <path>",
            )

    with open(path) as stream:
        raw_yaml = yaml.safe_load(stream)
        if environment not in raw_yaml:
            raise Exception(
                f"Environment '{environment}' not specified in configuration file at '{path}'",
            )
        data = raw_yaml[environment]

    for field in fields(UserCodeDeploymentsConfig):
        if os.environ.get(field.name.upper(), None) is not None:
            data[field.name] = os.environ[field.name.upper()]

    logger.debug(f"Using configuration:\n {data}")
    return UserCodeDeploymentsConfig(**data)



# This is a k8s yaml for defining a custom configmap with yaml data in it, loaded into a dictionary.
# We will put the values.yaml for dagster's user-code deployment chart inside, and thereby use it to keep track of
# currently deployed user code deployments and their configuration *on the k8s cluster*.
BASE_CONFIGMAP = {
    "apiVersion": "v1",
    "kind": "ConfigMap",
    "data": {"yaml": ""},
    "metadata": {},
}

# This is the data we want to insert into our custom configmap. It's the 'values.yaml' that is to be fed into the
# dagster user-code deployment chart later. This values.yaml starts out with an empty deployment list and is later
# modified and updated to contain all the relevant deployments and credentials.
BASE_CONFIGMAP_DATA = {
    "global": {
        "postgresqlSecretName": "",
        "dagsterHome": "",
        "serviceAccountName": "",
        "celeryConfigSecretName": "",
    },
    "dagsterHome": "/opt/dagster/dagster_home",
    "postgresqlSecretName": "dagster-postgresql-secret",
    "celeryConfigSecretName": "dagster-celery-config-secret",
    "deployments": [],
    "imagePullSecrets": [],
    "serviceAccount": {"create": True, "name": "", "annotations": {}},
    "rbacEnabled": True,
    "extraManifests": [],
}



import logging
import sys

logging.basicConfig(
    level=logging.WARNING,
    handlers=[logging.StreamHandler(sys.stdout)],
    format="%(asctime)s - %(name)s.%(funcName)s():%(lineno)d - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger(__name__)



import os
import re
import subprocess
import sys
from enum import Enum
from subprocess import Popen, TimeoutExpired

import typer

from dagster_uc.log import logger


class BuildTool(str, Enum):
    """The possible build tools to choose from"""

    podman = "podman"
    docker = "docker"
    auto = "auto"


def exception_on_failed_subprocess(res: subprocess.CompletedProcess) -> None:
    """Raises a python exception if the subprocess terminated with exit code > 0"""
    if res.stdout is not None:
        logger.debug(res.stdout.decode("utf-8"))
    if res.returncode != 0:
        if res.stderr is not None:
            logger.error(res.stderr.decode("utf-8"))
        raise Exception("Subprocess failed")


def run_cli_command(
    cmd: str,
    ignore_failures: bool = False,
    input_str: str | None = None,
    capture_output: bool = False,
    timeout: int | None = None,
    shell: bool = True,
) -> subprocess.CompletedProcess | TimeoutExpired:
    """Run the cli command while capturing the output, so its output is not directly shown."""
    logger.debug(f"[running command] {cmd}")
    if input_str == "":
        input_str_bytes = None
    elif input_str is not None:
        input_str_bytes = input_str.encode("utf-8")
    else:
        input_str_bytes = None
    try:
        res = subprocess.run(
            cmd,
            shell=shell,
            check=False,
            env=os.environ.copy(),
            capture_output=capture_output,
            input=input_str_bytes,
            timeout=timeout,
        )
    except TimeoutExpired as timeout_expired:
        return timeout_expired
    if not ignore_failures:
        exception_on_failed_subprocess(res)
    return res


def gen_tag(
    deployment_name: str,
    container_registry: str,
    dagster_version: str,
    use_az_login: bool,
) -> str:
    """Identifies the latest tag present in the ACR and increments it by one."""
    if use_az_login:
        login_registry(container_registry)

    res = run_cli_command(
        f"podman search {os.path.join(container_registry, deployment_name)} --list-tags --format {{{{.Tag}}}} --limit 9999999",
        ignore_failures=True,
        capture_output=True,
        timeout=15,
    )
    if isinstance(res, TimeoutExpired):
        if res.stderr is not None and "Please try running 'az login' again" in res.stderr.decode():
            raise Exception(
                f"{res.stderr.decode()}\n\nAzure cli is likely not logged in. Please try 'az login'",
            )
        else:
            raise res
    if res.stderr is not None and any(
        txt in res.stderr.decode("utf-8").lower() for txt in ["is not found", "does not exist"]
    ):
        return f"{dagster_version}-0"
    elif res.returncode > 0:
        raise Exception(res.stderr)

    tags = re.findall(r".*?-(\d+)", res.stdout.decode("utf-8"))
    logger.debug(
        f"Found the following image tags for this branch in the container registry: {tags}",
    )

    tags_ints = [int(tag) for tag in tags]
    if not len(tags_ints):
        return f"{dagster_version}-0"
    new_tag = f"{dagster_version}-{max(tags_ints) + 1}"
    return new_tag


def run_cli_command_streaming(cmd: str, as_user: str = "") -> None:
    """Run the cli command without capturing the output, so it streams to the console."""
    if as_user:
        cmd = f"sudo -u {as_user} {cmd}"
    logger.debug(f"[running command] {cmd}")
    Popen(cmd, stdout=sys.stdout, stderr=sys.stderr, env=os.environ.copy(), shell=True)


def login_registry(image_registry: str) -> None:
    """Logs into registry with az cli"""
    typer.echo("Logging into acr...")
    cmd = [
        "az",
        "acr",
        "login",
        "--name",
        image_registry,
        "--expose-token",
        "--output",
        "tsv",
        "--query",
        "accessToken",
    ]
    typer.echo(f"\033[1mExecuting\033[0m cmd: {' '.join(cmd)}")
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=False)
    output, error = process.communicate()
    token = output.decode().strip().encode("utf-8")

    cmd = [
        BuildTool.podman.value,
        "login",
        "--username",
        "00000000-0000-0000-0000-000000000000",
        "--password-stdin",
        image_registry,
    ]
    exception_on_failed_subprocess(subprocess.run(cmd, input=token, capture_output=False))


def build_and_push(
    repository_root: str,
    image_registry: str,
    image_name: str,
    dockerfile: str,
    use_sudo: bool,
    tag: str,
    branch_name: str,
    use_az_login: bool,
):
    """Build a docker image and push it to the registry"""
    # We need to work from the root of the repo so docker can access all files
    previous_dir = os.getcwd()
    os.chdir(repository_root)

    cmd = [
        BuildTool.podman.value,
        "build",
        "-f",
        os.path.join(os.getcwd(), dockerfile),
        "-t",
        os.path.join(image_registry, f"{image_name}:{tag}"),
        "--build-arg=BRANCH_NAME=" + branch_name,
        ".",
    ]

    if use_sudo:
        cmd = ["sudo"] + cmd

    exception_on_failed_subprocess(subprocess.run(cmd, capture_output=False))

    if use_az_login:
        login_registry(image_registry=image_registry)

    typer.echo("Pushing image...")
    cmd = [BuildTool.podman.value, "push", os.path.join(image_registry, f"{image_name}:{tag}")]
    if use_sudo:
        cmd = ["sudo"] + cmd
    exception_on_failed_subprocess(subprocess.run(cmd, capture_output=False))
    os.chdir(previous_dir)



import asyncio
import logging
import re
import subprocess
from collections.abc import Callable
from typing import NamedTuple

import kr8s
import yaml
from kr8s.objects import (
    ConfigMap,
    Deployment,
    Pod,
)
from kr8s.objects import Role as Role
from kr8s.objects import RoleBinding as RoleBinding
from kr8s.objects import Service as Service
from kr8s.objects import ServiceAccount as ServiceAccount
from packaging.version import Version

from dagster_uc.config import UserCodeDeploymentsConfig
from dagster_uc.configmaps import BASE_CONFIGMAP, BASE_CONFIGMAP_DATA
from dagster_uc.log import logger


class DagsterDeployment(NamedTuple):
    """Dagster deployment names"""

    full_name: str
    branch_name: str


class DagsterUserCodeHandler:
    """This the dagster-user code handler for common activities such as updating config maps, listing them and modifying them."""

    def __init__(self, config: UserCodeDeploymentsConfig, kr8s_api: kr8s.Api) -> None:
        self.config = config
        self.api = kr8s_api

    def maybe_create_user_deployments_configmap(self) -> None:
        """Creates a user deployments_configmap if it doesn't exist yet."""
        from copy import deepcopy

        dagster_user_deployments_values_yaml_configmap = deepcopy(BASE_CONFIGMAP)
        dagster_user_deployments_values_yaml_configmap["metadata"]["name"] = (
            self.config.user_code_deployments_configmap_name
        )
        dagster_user_deployments_values_yaml_configmap["data"]["yaml"] = yaml.dump(
            BASE_CONFIGMAP_DATA,
        )
        try:
            self._read_namespaced_config_map(
                self.config.user_code_deployments_configmap_name,
            )
        except kr8s.NotFoundError:
            ConfigMap(
                resource=dagster_user_deployments_values_yaml_configmap,
                namespace=self.config.namespace,
                api=self.api,
            ).create()

    def remove_all_deployments(self) -> None:
        """This function removes in its entirety the values.yaml for dagster's user-code deployment chart from the k8s
        cluster and replaces it with one with an empty deployments array as read
        from dagster_user_deployments_values_yaml_configmap.
        """
        from copy import deepcopy

        dagster_user_deployments_values_yaml_configmap = deepcopy(BASE_CONFIGMAP)
        dagster_user_deployments_values_yaml_configmap["data"]["yaml"] = yaml.dump(
            BASE_CONFIGMAP_DATA,
        )

        configmap = self._read_namespaced_config_map(
            self.config.user_code_deployments_configmap_name,
        )
        configmap.patch(dagster_user_deployments_values_yaml_configmap)

    def list_deployments(
        self,
    ) -> list[dict]:
        """Get the contents of the deployments array from the values.yaml of dagster's user-code deployment chart as it is
        currently stored on k8s.
        """
        config_map = self._read_namespaced_config_map(
            self.config.user_code_deployments_configmap_name,
        )
        current_deployments: list = yaml.safe_load(config_map["data"]["yaml"])["deployments"]
        return current_deployments

    def get_deployment(
        self,
        name: str,
    ) -> dict | None:
        """Return None if the deployment does not exist. Otherwise, return the deployment config."""
        current_deployments = self.list_deployments()
        deployments = list(filter(lambda x: x["name"] == name, current_deployments))
        if len(deployments):
            return deployments[0]
        else:
            return None

    def _check_deployment_exists(
        self,
        name: str,
    ) -> bool:
        """Return True if the deployment exists. This is done by reading the configmap of values.yaml for dagster's
        user-code deployment chart and checking if the deployments array contains this particular deployment_name
        """
        return self.get_deployment(name) is not None

    def update_dagster_workspace_yaml(
        self,
    ) -> None:
        """This function updates dagster's dagster-workspace-yaml configmap to include all currently configured
        deployments. Should be called after adding or removing user-code deployments.
        """
        configmap = self._read_namespaced_config_map(
            self.config.dagster_workspace_yaml_configmap_name,
        )

        last_applied_configuration = (
            configmap["metadata"]
            .get("annotations", {})
            .get("kubectl.kubernetes.io/last-applied-configuration", None)
        )

        def generate_grpc_servers_yaml(servers: list[dict]) -> str:
            data = {"load_from": []}
            for server in servers:
                grpc_server = {
                    "host": server["name"],
                    "port": 3030,
                    "location_name": server["name"].replace(
                        "--",
                        ":",
                    ),  ## We replace the -- separator with `:` for more friendly UI name
                }
                data["load_from"].append({"grpc_server": grpc_server})
            return yaml.dump(data)

        workspaceyaml = generate_grpc_servers_yaml(
            self.list_deployments(),
        )

        new_configmap = {
            "apiVersion": "v1",
            "kind": "ConfigMap",
            "data": {"workspace.yaml": workspaceyaml},
        }
        new_configmap["metadata"] = {
            "name": self.config.dagster_workspace_yaml_configmap_name,
            "namespace": self.config.namespace,
            "annotations": {
                "kubectl.kubernetes.io/last-applied-configuration": last_applied_configuration,
            },
        }
        configmap.patch(new_configmap)  # type: ignore

    def deploy_to_k8s(
        self,
        reload_dagster: bool = True,
    ) -> None:
        """This will read the values.yaml for dagster's user-code deployment chart as it exists on k8s, and feed it into
        dagster's user-code deployment chart, generating k8s yaml files for user-code deployments. These yamls are
        applied such that the cluster will now reflect the latest version of the user-code deployment configuration.
        """
        from datetime import datetime

        from pyhelm3 import Client
        from pytz import timezone

        tz = timezone("Europe/Amsterdam")

        values_dict = yaml.safe_load(
            self._read_namespaced_config_map(self.config.user_code_deployments_configmap_name)[
                "data"
            ]["yaml"],
        )
        self.update_dagster_workspace_yaml()

        loop = asyncio.new_event_loop()
        helm_client = Client()
        chart = loop.run_until_complete(
            helm_client.get_chart(
                chart_ref="dagster-user-deployments",
                repo="https://dagster-io.github.io/helm",
                version=self.config.dagster_version,
            ),
        )
        helm_templates = [
            *loop.run_until_complete(
                helm_client.template_resources(
                    chart,
                    "dagster",
                    values_dict,
                    namespace=self.config.namespace,
                ),
            ),
        ]

        # Update user code deployments in k8s (akin to kubectl apply -f)
        for obj in helm_templates:
            k8s_obj = eval(obj["kind"])(obj, api=self.api)
            try:
                k8s_obj.patch(obj)
            except kr8s.NotFoundError:
                k8s_obj.create()

        if reload_dagster:
            for deployment_name in ["dagster-daemon", "dagster-dagster-webserver"]:
                deployment = Deployment.get(deployment_name, namespace=self.config.namespace)
                reload_patch = {
                    "spec": {
                        "template": {
                            "metadata": {
                                "annotations": {
                                    "kubectl.kubernetes.io/restartedAt": datetime.now(tz).strftime(
                                        "%Y-%m-%dT%H:%M:%S%z",
                                    ),
                                },
                            },
                        },
                    },
                }
                deployment.patch(reload_patch)

    def delete_k8s_resources_for_user_deployment(
        self,
        label: str,
        delete_deployments: bool = True,
    ) -> None:
        """Deletes all k8s resources related to a specific user code deployment.
        Returns a boolean letting you know if pod was found
        """
        for pod in self.api.get(
            Pod,
            label_selector=f"dagster/code-location={label}",
            field_selector="status.phase=Succeeded",
            namespace=self.config.namespace,
        ):
            logger.info(f"Deleting pod {pod.name}")
            pod.delete()

        if delete_deployments:
            import contextlib

            with contextlib.suppress(kr8s.NotFoundError):
                Deployment.get(
                    namespace=self.config.namespace,
                    label_selector=f"deployment={label}",
                    api=self.api,
                ).delete()
                Deployment.get(
                    namespace=self.config.namespace,
                    label_selector=f"dagster/code-location={label}",
                    api=self.api,
                ).delete()

    def gen_new_deployment_yaml(
        self,
        name: str,
        image_prefix: str | None,
        tag: str,
    ) -> dict:
        """This function generates yaml for a single user-code deployment, which is to be part of the 'deployments' array in the
        values.yaml for dagster's user-code deployments chart.
        """
        import os

        deployment = {
            "name": name,
            "image": {
                "repository": os.path.join(
                    self.config.container_registry,
                    image_prefix or "",
                    name,
                ),
                "tag": tag,
                "pullPolicy": "Always",
            },
            "dagsterApiGrpcArgs": [
                "-f",
                os.path.join(self.config.docker_root, self.config.code_path),
            ],
            "port": 3030,
            "includeConfigInLaunchedRuns": {"enabled": True},
            "env": self.config.user_code_deployment_env,
            "envConfigMaps": [],
            "envSecrets": self.config.user_code_deployment_env_secrets,
            "annotations": {},
            "nodeSelector": {"agentpool": self.config.node},
            "affinity": {},
            "resources": {
                "limits": self.config.limits,
                "requests": self.config.requests,
            },
            "tolerations": [
                {
                    "key": "agentpool",
                    "operator": "Equal",
                    "value": self.config.node,
                    "effect": "NoSchedule",
                },
            ],
            "podSecurityContext": {},
            "securityContext": {},
            "labels": {},
            "readinessProbe": {
                "enabled": True,
                "periodSeconds": 20,
                "timeoutSeconds": 10,
                "successThreshold": 1,
                "failureThreshold": 3,
            },
            "livenessProbe": {},
            "startupProbe": {"enabled": False},
            "service": {"annotations": {}},
        }
        logger.debug(f"Generated user code deployment:\n{deployment}")
        return deployment

    def _read_namespaced_config_map(
        self,
        name: str,
    ) -> ConfigMap:
        """Read a configmap that exists on the k8s cluster"""
        configmap = ConfigMap.get(name=name, namespace=self.config.namespace, api=self.api)
        return configmap

    def add_user_deployment_to_configmap(
        self,
        new_deployment: dict,
    ) -> None:
        """This function takes a new user-code deployment yaml and adds it to the deployments array
        in the values.yaml of dagster's user-code deployment chart.
        (referring to the values.yaml that is stored in a configmap on k8s.)
        """

        def modify_func(current_deployments: list[dict]) -> list[dict]:
            return current_deployments + [new_deployment]

        self._modify_user_deployments(modify_func)

    def remove_user_deployment_from_configmap(
        self,
        name: str,
    ) -> None:
        """This function removes a user-code deployment yaml from the deployments array
        in the values.yaml of dagster's user-code deployment chart.
        (referring to the values.yaml that is stored in a configmap on k8s.)
        """

        def modify_func(current_deployments: list[dict]) -> list[dict]:
            filtered = list(filter(lambda d: d["name"] != name, current_deployments))
            if len(filtered) == len(current_deployments):
                logger.warning(
                    f'Deployment name "{name}" does not seem to exist in environment "{self.config.environment}". Proceeding to attempt deletion of k8s resources anyways.',
                )
            return filtered

        self._modify_user_deployments(modify_func)

    def _modify_user_deployments(
        self,
        modify_func: Callable[[list[dict]], list[dict]],
    ) -> None:
        """Modifies the deployments array of the values.yaml for Dagster's user-code deployment chart on k8s.

        This function allows for customization of the deployments array by providing a `modify_func` which
        will process the current list of deployments and should return the modified list of deployments.
        This operation is treated as a transaction.

        Args:
            modify_func (Callable[[List[dict]], List[dict]]): A function that takes the current list of
                deployments as input and returns the modified list of deployments.
            config (UserCodeDeploymentsConfig): Config object
        Examples:
            To keep only the first deployment, you can pass the following `modify_func`

            >>> modify_user_deployments(lambda deployment_list: deployment_list[0:1])
        """
        from copy import deepcopy

        configmap = self._read_namespaced_config_map(
            self.config.user_code_deployments_configmap_name,
        )
        last_applied_configuration = (
            configmap["metadata"]
            .get("annotations", {})
            .get("kubectl.kubernetes.io/last-applied-configuration", None)
        )
        current_deployments: list = yaml.safe_load(configmap["data"]["yaml"])["deployments"]

        current_deployments = modify_func(current_deployments)

        depl_list_str = (
            "\n".join([d["name"] for d in current_deployments])
            if len(current_deployments)
            else "No deployments"
        )
        logging.debug(f"List of currently configured deployments:\n{depl_list_str}\n\n")
        new_configmap_data = deepcopy(BASE_CONFIGMAP_DATA)
        new_configmap_data["deployments"] = current_deployments

        new_configmap = deepcopy(BASE_CONFIGMAP)
        new_configmap["data"]["yaml"] = yaml.dump(new_configmap_data)

        new_configmap["metadata"] = {
            "name": self.config.user_code_deployments_configmap_name,
            "namespace": self.config.namespace,
            "annotations": {
                "kubectl.kubernetes.io/last-applied-configuration": last_applied_configuration,
            },
        }
        configmap.patch(new_configmap)

    def get_deployment_name(  # noqa: D102
        self,
        deployment_name_suffix: str | None = None,
        use_project_name: bool = True,
    ) -> DagsterDeployment:
        """Creates a deployment name based on the name of the pyproject.toml and name of git branch"""
        logger.debug("Determining deployment name...")

        project_name = self._get_project_name() if use_project_name else None

        if not self.config.cicd:
            branch = subprocess.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"]).decode()
            if deployment_name_suffix:
                branch += deployment_name_suffix
            branch = re.sub(r"[^a-zA-Z0-9]+", "-", branch).strip("-")  # Strips double --
            name = f"{project_name}--{branch}" if project_name is not None else branch

        else:
            branch = self.config.environment
            name = f"{project_name}--{branch}" if project_name is not None else branch
        return DagsterDeployment(
            full_name=name,
            branch_name=branch,
        )

    def _ensure_dagster_version_match(self) -> None:
        """Raises an exception if the cluster version of dagster is different than the local version"""
        logger.debug("Going to read the cluster dagster version...")
        local_dagster_version = Version(self.config.dagster_version)

        ## GETS cluster version from dagster deamon pod
        deamon_pod = Pod.get(
            label_selector="deployment=daemon",
            namespace=self.config.namespace,
            api=self.api,
        )

        ex = deamon_pod.exec(command=["dagster", "--version"])
        output = ex.stdout.decode("ascii")  # type: ignore
        cluster_dagster_version = re.findall("version (.*)", output)

        if len(cluster_dagster_version) != 1:
            raise Exception(
                f"Failed parsing the cluster dagster version, exec response from container `{output}`",
            )
        else:
            cluster_dagster_version = Version(cluster_dagster_version[0])

        logger.debug(f"Cluster dagster version detected to be '{cluster_dagster_version}'")
        if not cluster_dagster_version == local_dagster_version:
            raise Exception(
                f"Dagster version mismatch. Local: {local_dagster_version}, Cluster: {cluster_dagster_version}. Try pulling the latest changes from the develop branch and then rebuilding the local python environment.",
            )

    def check_if_code_pod_exists(self, label: str) -> bool:
        """Checks if the code location pod of specific label is available"""
        running_pods = list(
            self.api.get(
                Pod,
                label_selector=f"deployment={label}",
                namespace=self.config.namespace,
            ),
        )
        return len(running_pods) > 0

    def delete_k8s_resources(self, label_selector: str):
        """Delete all k8s resources with a specified label_selector"""
        for resource in [
            "Pod",
            "ReplicationController",
            "Service",
            "DaemonSet",
            "Deployment",
            "ReplicaSet",
            "StatefulSet",
            "HorizontalPodAutoscaler",
            "CronJob",
            "Job",
        ]:
            for item in self.api.get(
                resource,
                namespace=self.config.namespace,
                label_selector=label_selector,
            ):
                item.delete()

    def acquire_semaphore(self, reset_lock: bool = False) -> bool:
        """Acquires a semaphore by creating a configmap"""
        if reset_lock:
            try:
                semaphore = ConfigMap.get(
                    self.config.uc_deployment_semaphore_name,
                    namespace=self.config.namespace,
                    api=self.api,
                )
                semaphore.delete()
            except kr8s.NotFoundError:
                pass

        try:
            semaphore = ConfigMap.get(
                self.config.uc_deployment_semaphore_name,
                namespace=self.config.namespace,
                api=self.api,
            )
            if semaphore.data.get("locked") == "true":
                return False
            else:
                semaphore.patch({"data": {"locked": "true"}})
                return True
        except kr8s.NotFoundError:
            # Create semaphore if it does not exist
            semaphore = ConfigMap(
                {
                    "metadata": {
                        "name": self.config.uc_deployment_semaphore_name,
                        "namespace": self.config.namespace,
                    },
                    "data": {"locked": "true"},
                },
                api=self.api,
            ).create()
            return True

    def release_semaphore(self) -> None:
        """Releases the semaphore lock"""
        try:
            semaphore = ConfigMap.get(
                self.config.uc_deployment_semaphore_name,
                namespace=self.config.namespace,
                api=self.api,
            )
            semaphore.patch({"data": {"locked": "false"}})
            logger.debug("patched semaphore to locked: false")
        except Exception as e:
            logger.error(f"Failed to release deployment lock: {e}")

    def _get_project_name(self) -> str | None:
        import tomli

        try:
            with open("pyproject.toml", "rb") as fp:
                data = tomli.load(fp)
                return re.sub("[^a-zA-Z0-9-]", "-", data["project"]["name"]).strip("-")
        except FileNotFoundError:
            logger.warning("""
            pyproject.toml not found, no project name will be used.
            Make sure dagster-uc is called in the same directory as the project""")
            return None



#!/usr/bin/env python3
# ruff: noqa: D103
import contextlib
import json
import logging
import os
import pprint
import subprocess
import time
from dataclasses import asdict
from typing import Annotated, cast

import kr8s
import typer
from kr8s.objects import (
    ConfigMap,
    Pod,
)

from dagster_uc.config import UserCodeDeploymentsConfig, load_config
from dagster_uc.log import logger
from dagster_uc.uc_handler import DagsterUserCodeHandler
from dagster_uc.utils import BuildTool, build_and_push, gen_tag

app = typer.Typer(invoke_without_command=True)
deployment_app = typer.Typer(
    name="deployment",
    help="Contains various subcommands for managing user code deployments",
    no_args_is_help=True,
)
app.add_typer(deployment_app)
deployment_delete_app = typer.Typer(
    name="delete",
    help="Delete one or more user code deployments",
)
deployment_app.add_typer(deployment_delete_app)
deployment_check_app = typer.Typer(
    name="check",
    help="Check status of the user code deployment",
)
deployment_app.add_typer(deployment_check_app)

handler: DagsterUserCodeHandler
config: UserCodeDeploymentsConfig


@app.command("show-config", help="Outputs the configuration that is currently in use")
def show_config():
    """Pretty print the config object"""
    pprint.pprint(config, indent=4)


@app.callback(invoke_without_command=True, no_args_is_help=True)
def default(
    ctx: typer.Context,
    environment: str = typer.Option("dev", "--environment", "-e", help="The environment"),
    config_file_path: str = typer.Option(
        None,
        "--config-file",
        "-c",
        help="Path to the config file.",
    ),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Output DEBUG logging"),
) -> None:
    """This function executes before any other nested cli command is called and loads the configuration object."""
    global logger
    global config
    global handler

    logger.setLevel(logging.DEBUG if verbose else logging.INFO)
    config = load_config(environment, config_file_path)

    if ctx.invoked_subcommand == "init-config":
        pass
    else:
        if verbose:
            config.verbose = True
        logger.debug(f"Switching kubernetes context to {config.environment}...")
        kr8s_api = kr8s.api(context=f"{config.kubernetes_context}", namespace=config.namespace)

        handler = DagsterUserCodeHandler(config, kr8s_api)
        handler._ensure_dagster_version_match()
        handler.maybe_create_user_deployments_configmap()
        logger.debug(f"Done: Switched kubernetes context to {config.environment}")


def build_push_container(
    deployment_name: str,
    branch_name: str,
    image_prefix: str | None,
    config: UserCodeDeploymentsConfig,
    use_sudo: bool,
    tag: str,
) -> None:
    """Builds a docker image for a user-code deployment of the current branch and uploads it to the image registry"""
    handler.update_dagster_workspace_yaml()
    build_and_push(
        config.repository_root,
        config.container_registry,
        image_name=deployment_name
        if not image_prefix
        else os.path.join(image_prefix, deployment_name),
        dockerfile=config.dockerfile,
        use_sudo=use_sudo,
        tag=tag,
        branch_name=branch_name,
        use_az_login=config.use_az_login,
    )


@app.command(
    name="init-config",
    help="Create a config for dagster deployment",
    no_args_is_help=True,
)
def init_config(
    file: Annotated[
        str,
        typer.Option(
            "--file",
            "-f",
            help="File and path where to save the config example.",
        ),
    ],
):
    """Initiates a config template"""

    def optional_prompt(text: str) -> str | None:
        var = typer.prompt(text, default="")
        if var:
            return var
        else:
            return None

    initialized_config = UserCodeDeploymentsConfig(
        environment=typer.prompt("""What environment is this config for [dev, acc, prod etc.]"""),
        container_registry=typer.prompt("Container registry address"),
        dockerfile=typer.prompt("Path of dockerfile", default="./Dockerfile"),
        image_prefix=typer.prompt("Prefix for container image to use"),
        namespace=typer.prompt("Namespace of kubernetes deployment"),
        node=typer.prompt("Kubernetes node for user-code pod"),
        code_path=typer.prompt("Path of the Definitions python file inside docker image"),
        docker_root=typer.prompt("Path of docker scope", default="."),
        repository_root=typer.prompt("Path of project scope", default="."),
        dagster_version=typer.prompt(
            "Version of dagster in project (should mirror dagster depoyment!)",
        ),
        user_code_deployment_env_secrets=[],
        user_code_deployment_env=[],
        cicd=typer.confirm(
            "Whether it's executed in CICD. If set to True, then the deployment_name is created from the env",
        ),
        requests=json.loads(
            typer.prompt(
                "Request for the user pod in k8s in json formatted string",
                default=json.dumps({"cpu": "1", "memory": "1Gi"}),
            ),
        ),
        limits=json.loads(
            typer.prompt(
                "Limits for the user pod in k8s in json formatted string",
                default=json.dumps({"cpu": "2", "memory": "2Gi"}),
            ),
        ),
        kubernetes_context=typer.prompt("Kubernetes context of the cluster to use for api calls"),
        dagster_gui_url=optional_prompt("URL of dagster UI"),
        use_az_login=typer.confirm("Whether to use az cli to login to container registry"),
        use_project_name=typer.confirm(
            "Whether to use the pyproject.toml project-name as deployment name prefix.",
        ),
        user_code_deployments_configmap_name=typer.prompt(
            "Configmap name to use for user_code_deployments",
            default="dagster-user-deployments-values-yaml",
        ),
        dagster_workspace_yaml_configmap_name=typer.prompt(
            "Configmap name of the dagster_workspace_yaml",
            default="dagster-workspace-yaml",
        ),
    )
    with open(file, "w") as fp:
        import yaml

        config_dict = asdict(initialized_config)
        complete_dict = {config_dict["environment"]: config_dict}
        yaml.dump(complete_dict, fp, default_flow_style=False)
    typer.echo(f"Template configuration file generated as '{file}'.")


@deployment_app.command(
    name="list",
    help="List user code deployments that are currently active",
)
def deployment_list():
    """Outputs a list of currently active deployments"""
    typer.echo(
        "\033[1mActive user code deployments\033[0m\n"
        + "\n".join(
            ["* " + d["name"] for d in handler.list_deployments()],
        ),
    )


@deployment_app.command(
    name="revive",
    help="Redeploy an old user code deployment using the latest existing image from the container registry",
)
def deployment_revive(
    name: Annotated[
        str,
        typer.Option("--name", "-n", help="The name of the deployment to revive."),
    ],
    tag: Annotated[str, typer.Option("--tag", "-t", help="The tag of the deployment to revive.")],
):
    # In case the UI name separator of the deployment is passed
    name = name.replace(":", "--")

    if not handler._check_deployment_exists(
        name,
    ):
        handler.add_user_deployment_to_configmap(
            handler.gen_new_deployment_yaml(
                name,
                image_prefix=handler.config.image_prefix,
                tag=tag,
            ),
        )
        handler.deploy_to_k8s()
        logger.info(f"Deployed {name}")
    else:
        raise Exception(f'Deployment "{name}" already exists')


@deployment_delete_app.callback(invoke_without_command=True)
def deployment_delete(
    delete_all: Annotated[
        bool,
        typer.Option(
            "--all",
            "-a",
            help="If this is provided, all deployments will be deleted including CI/CD provided deployments.",
        ),
    ] = False,
    name: Annotated[
        str,
        typer.Option(
            "--name",
            "-n",
            help="The name of the deployment to delete. If none is provided, the deployment corresponding to the currently checked out git branch will be deleted.",
        ),
    ] = "",
) -> None:
    if delete_all:
        handler.remove_all_deployments()
        handler.delete_k8s_resources(
            label_selector="app.kubernetes.io/name=dagster-user-deployments",
        )
        handler.delete_k8s_resources(label_selector="app=dagster-user-deployments")
        for item in handler.api.get(
            ConfigMap,
            namespace=config.namespace,
            label_selector="app=dagster-user-deployments",
        ):
            item.delete()  # type: ignore
        handler.delete_k8s_resources(label_selector="dagster/code-location")
        handler.deploy_to_k8s()
        typer.echo("\033[1mDeleted all deployments\033[0m")
    else:
        if not name:
            name = handler.get_deployment_name(
                deployment_name_suffix="",
                use_project_name=config.use_project_name,
            ).full_name
        else:
            # In case the UI name separator of the deployment is passed
            name = name.replace(":", "--")
        handler.remove_user_deployment_from_configmap(name)
        handler.delete_k8s_resources_for_user_deployment(
            name,
            delete_deployments=True,
        )
        handler.deploy_to_k8s(reload_dagster=True)
        typer.echo(f"Deleted deployment \033[1m{name}\033[0m")


@deployment_check_app.callback(invoke_without_command=True)
def check_deployment(
    name: Annotated[
        str,
        typer.Option(
            "--name",
            "-n",
            help="The name of the deployment to check. If not provided, checks deployment corresponding to the current branch.",
        ),
    ] = "",
    timeout: Annotated[
        int,
        typer.Option(
            "--timeout",
            "-t",
            help="The timeout duration in seconds to keep following the logs of user code pod.",
        ),
    ] = 60,
) -> None:
    """This function executes before any other nested cli command is called and loads the configuration object."""
    if not name:
        name = handler.get_deployment_name(use_project_name=config.use_project_name).full_name
    else:
        # In case the UI name separator of the deployment is passed
        name = name.replace(":", "--")
    if not handler._check_deployment_exists(name):
        logger.warning(
            f"Deployment with name '{name}' does not seem to exist in environment '{config.environment}'. Attempting to proceed with status check anyways.",
        )
    typer.echo(f"\033[1mStatus for deployment {name}\033[0m")
    for pod in handler.api.get(
        Pod,
        label_selector=f"deployment={name}",
        namespace=config.namespace,
    ):
        pod = cast(Pod, pod)
        with contextlib.suppress(Exception):
            for line in pod.logs(pretty=True, follow=True, timeout=timeout):  # type: ignore
                typer.echo(line)
                if "started dagster code server" in line.lower():
                    break


@deployment_app.command(
    name="deploy",
    help="Deploys the currently checked out git branch as a user code deployment",
)
def deployment_deploy(
    force: Annotated[
        bool,
        typer.Option(
            "--force",
            "-f",
            help="If this is provided, a full redeploy will always be done, rather than just rebooting user code pods if they already exist in order to trigger a new image pull",
        ),
    ] = False,
    skip_build: bool = typer.Option(
        False,
        "--skip-build",
        "-b",
        help="Image build and push will be skipped",
    ),
    deployment_name_suffix: str = typer.Option(
        "",
        "--deployment-name-suffix",
        "-s",
        help="Suffix to append to the default deployment name (which is based on the git branch name) - useful for doing parallel deployments from the same branch.",
    ),
    deployment_name: str = typer.Option(
        "",
        "--deployment-name",
        "-n",
        help="Overrides the name of the deployment, including any --deployment-name-suffix value",
    ),
    reset_lock: bool = typer.Option(
        False,
        "--reset-lock",
        "-r",
        help="Reset the deployment semaphore of any ongoing other deployments.",
    ),
    use_sudo: Annotated[
        bool,
        typer.Option(
            "--use-sudo",
            "-u",
            help="If this is provided, buildah or docker will be called with sudo",
        ),
    ] = False,
):
    def is_command_available(command: str) -> bool:
        try:
            subprocess.run(
                [command, "--version"],
                capture_output=True,
                check=True,  # ruff: ignore
            )
            return True
        except subprocess.CalledProcessError:
            return False
        except FileNotFoundError:
            return False

    count = 0
    while not handler.acquire_semaphore(reset_lock):
        logger.error(
            f"Attempt {count}: Another deployment is in progress. Trying again in 10 seconds. You can force a reset of the deployment lock by using 'dagster-uc deployment deploy --reset-lock'",
        )
        count += 1
        time.sleep(10)

    try:
        logger.debug("Determining build tool...")
        if not is_command_available(BuildTool.podman.value):
            raise Exception("Podman installation is required to run dagster-uc.")

        logger.debug("Using 'podman' to build image.")
        if deployment_name:
            (deployment_name, branch_name) = (deployment_name, deployment_name)
        else:
            dagster_deployment = handler.get_deployment_name(
                deployment_name_suffix,
                use_project_name=config.use_project_name,
            )
            (deployment_name, branch_name) = (
                dagster_deployment.full_name,
                dagster_deployment.branch_name,
            )

        logger.debug("Determining tag...")
        new_tag = gen_tag(
            deployment_name
            if not config.image_prefix
            else os.path.join(config.image_prefix, deployment_name),
            config.container_registry,
            config.dagster_version,
            config.use_az_login,
        )

        typer.echo(f"Deploying deployment \033[1m'{deployment_name}:{new_tag}'\033[0m")

        full_redeploy_done = False
        if not skip_build:
            build_push_container(
                deployment_name,
                branch_name=branch_name,
                image_prefix=handler.config.image_prefix,
                config=config,
                use_sudo=use_sudo,
                tag=new_tag,
            )

        if not handler._check_deployment_exists(deployment_name):
            logger.info(
                f"Deployment with name '{deployment_name}' does not exist yet in '{config.environment}'. Adding deployment to configmap",
            )
            handler.add_user_deployment_to_configmap(
                handler.gen_new_deployment_yaml(
                    deployment_name,
                    image_prefix=handler.config.image_prefix,
                    tag=new_tag,
                ),
            )
            handler.deploy_to_k8s()
        else:
            logger.info(
                f"Deployment with name '{deployment_name}' exists in '{config.environment}'. Updating deployment in configmap",
            )
            handler.remove_user_deployment_from_configmap(deployment_name)
            handler.add_user_deployment_to_configmap(
                handler.gen_new_deployment_yaml(
                    deployment_name,
                    image_prefix=handler.config.image_prefix,
                    tag=new_tag,
                ),
            )
            if config.cicd or force:
                handler.delete_k8s_resources_for_user_deployment(deployment_name)
                handler.deploy_to_k8s()
            elif not handler.check_if_code_pod_exists(label=deployment_name):
                logger.info(
                    "Code deployment present in configmap but pod not found, triggering full deploy...",
                )
                handler.delete_k8s_resources_for_user_deployment(deployment_name, True)
                handler.deploy_to_k8s()  # Something went wrong - redeploy yamls and reload webserver
            else:
                logger.info(
                    "Code deployment present in configmap and pod found...",
                )
                handler.delete_k8s_resources_for_user_deployment(deployment_name, False)
                handler.deploy_to_k8s(reload_dagster=False)
    finally:
        handler.release_semaphore()
    if config.dagster_gui_url:
        typer.echo(
            f"Your assets: {config.dagster_gui_url.rstrip('/')}/locations/{deployment_name.replace('--', ':')}/assets\033[0m",
        )
    time.sleep(5)
    timeout = 40 if not full_redeploy_done else 240

    while True:
        code_pods = list(
            handler.api.get(
                Pod,
                label_selector=f"deployment={deployment_name}",
                namespace=config.namespace,
            ),
        )
        if len(code_pods) == 0:
            time.sleep(2)
            continue
        else:
            break

    with contextlib.suppress(Exception):
        for code_pod in code_pods:
            code_pod.wait("condition=Ready", timeout=timeout)  # type: ignore
    check_deployment(deployment_name)


if __name__ == "__main__":
    app()
